/*
  IM50175H 14-segment LCD driver
  Modified for GPS frequency standard
*/


#include <avr/io.h>


#define LATCH_DDR DDRB
#define LATCH_PORT PORTB
#define LATCH_BIT PB2
#define DATA_DDR DDRB
#define DATA_PORT PORTB
#define DATA_BIT PB1
#define CLK_DDR DDRB
#define CLK_PORT PORTB
#define CLK_BIT PB0


.text


send:
	ldi	r25, 4
1:	sbrc	r24, 0
	sbi	_SFR_IO_ADDR(DATA_PORT), DATA_BIT
	lsr	r24
	sbi	_SFR_IO_ADDR(CLK_PORT), CLK_BIT
	dec	r25
	cbi	_SFR_IO_ADDR(CLK_PORT), CLK_BIT
	cbi	_SFR_IO_ADDR(DATA_PORT), DATA_BIT
	brne	1b
	ret


.global	lcd_init
.func lcd_init
lcd_init:
	cbi	_SFR_IO_ADDR(LATCH_PORT), LATCH_BIT
	cbi	_SFR_IO_ADDR(DATA_PORT), DATA_BIT
	cbi	_SFR_IO_ADDR(CLK_PORT), CLK_BIT
	sbi	_SFR_IO_ADDR(LATCH_DDR), LATCH_BIT
	sbi	_SFR_IO_ADDR(DATA_DDR), DATA_BIT
	sbi	_SFR_IO_ADDR(CLK_DDR), CLK_BIT
	ret
.endfunc


.global	lcd_refresh
.func lcd_refresh
lcd_refresh:
	; 159-128
	ldi	XL,lo8(Lcd_buf)
	ldi	XH,hi8(Lcd_buf)
1:	ld	ZL, X+
	ldi	ZH, 0x00
	subi	ZL, lo8(-(tbl0 - 0x30))
	sbci	ZH, hi8(-(tbl0 - 0x30))
#ifdef __AVR_HAVE_LPMX__
	lpm	r24, Z
#else
	lpm
	mov	r24, r0
#endif
	rcall	send
	cpi	XL,lo8(Lcd_buf + 8)
	brne	1b
	; 127-120
	ld	r24, X
	lsl	r24
	rcall	send
	rcall	send
	; 119-88
2:	ld	ZL, -X
	ldi	ZH, 0x00
	subi	ZL, lo8(-(tbl0 - 0x30))
	sbci	ZH, hi8(-(tbl0 - 0x30))
#ifdef __AVR_HAVE_LPMX__
	lpm	r24, Z
#else
	lpm
	mov	r24, r0
#endif
	swap	r24
	rcall	send
	cpi	XL,lo8(Lcd_buf)
	brne	2b
	; 87-80
	ldi	r24, 0x00
	rcall	send
	rcall	send
	; 79-48
3:	ld	ZL, X+
	ldi	ZH, 0x00
	subi	ZL, lo8(-(tbl1 - 0x30))
	sbci	ZH, hi8(-(tbl1 - 0x30))
#ifdef __AVR_HAVE_LPMX__
	lpm	r24, Z
#else
	lpm
	mov	r24, r0
#endif
	rcall	send
	cpi	XL, lo8(Lcd_buf + 8)
	brne	3b
	; 47-40
	ld	r24, X
	rcall	send
	rcall	send
	; 39-8
4:	ld	ZL, -X
	ldi	ZH, 0x00
	subi	ZL, lo8(-(tbl1 - 0x30))
	sbci	ZH, hi8(-(tbl1 - 0x30))
#ifdef __AVR_HAVE_LPMX__
	lpm	r24, Z
#else
	lpm
	mov	r24, r0
#endif
	swap	r24
	rcall	send
	cpi	XL,lo8(Lcd_buf)
	brne	4b
	; 7-0
	rcall	send
	rcall	send

	sbi	_SFR_IO_ADDR(LATCH_PORT), LATCH_BIT
	cbi	_SFR_IO_ADDR(LATCH_PORT), LATCH_BIT

	ret
.endfunc


.section	.bss
.global	Lcd_buf
Lcd_buf:
	.byte	0
	.byte	0
	.byte	0
	.byte	0
	.byte	0
	.byte	0
	.byte	0
	.byte	0
	.byte	0


.section	.text
tbl0:
	.byte	0b00000000	; 0
	.byte	0b00000000	; 1
	.byte	0b00001001	; 2
	.byte	0b00001001	; 3
	.byte	0b00001001	; 4
	.byte	0b00001001	; 5
	.byte	0b00001001	; 6
	.byte	0b00000000	; 7
	.byte	0b00001001	; 8
	.byte	0b00001001	; 9
tbl1:
	.byte	0b11101101	; 0
	.byte	0b00101000	; 1
	.byte	0b01100101	; 2
	.byte	0b01101100	; 3
	.byte	0b10101000	; 4
	.byte	0b11001100	; 5
	.byte	0b11001101	; 6
	.byte	0b11101000	; 7
	.byte	0b11101101	; 8
	.byte	0b11101100	; 9
