/*
  GPS Frequency Standard
  for ATtiny2313

  T. Nakagawa

  2010-05-08
*/


#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include "usi_i2c.h"
#include "14seglcd.h"


#define BUF_SIZE 48	/* GPS͗pobt@TCY */
#define OSC_FRQ 10000000
#define OSC_TOL 20	/* IV[^̂̋e덷(2ppm) */
#define ADJ_K0 1
#define ADJ_K1 1
#define ADJ_LTHR 5
#define ADJ_LCNT 3
#define ADJ_HTHR 100
#define ADJ_HCNT 3
#define MOD_GPS 0
#define MOD_CAP 1


static uint8_t Gate = 0;		/* Q[g(0:1s, 1:10s, 2:100s) */
static uint8_t Gps_time[6] = "000000";	/* GPS̎ */
static uint8_t Gps_stt = 0;		/* GPS̎M */
static uint8_t Gps_sat = 0;		/* GPS̕ߑq */
static uint8_t Cap_new = 1;		/* TvO̊Jn\tO */
static uint16_t Cap_num;		/* ܂łɃJEgb */
static int16_t Cap_dif;			/* Q[gԓ̃IV[^̌덷̗ݐϒl */
static uint16_t Cap_err = 0;		/* IV[^̌덷̐Βl */
static uint32_t Cap_cnt;		/* 1ppsԂɃJEgNbN */
static uint8_t Cap_out;			/* Ol̃JEg */
static int16_t Adj_tune = 0;		/* VCTCXO̒l */
static int16_t Adj_temp;		/* VCTCXỎx */
volatile static uint8_t Modified = 0;		/* Ԃ̕ύX̗L */


static void adjust();
static void display();
static void monitor();
static void uart_putc(uint8_t chr);
static void uart_puts(uint8_t *str, uint8_t len);
static void uart_putnum(int16_t num);


int main(void) {
  /* foCX */
  ACSR |= _BV(ACD);	/* AiORp[^~ */
  MCUCR |= _BV(SE);	/* X[v */

  /* I2C */
  i2c_init();

  /* JE^ */
  TCCR1B |= _BV(CS10);
  TIMSK |= (_BV(TOIE1) | _BV(ICIE1));

  /* UART(4800baud) */
  UCSRB |= (_BV(RXCIE) | _BV(RXEN) | _BV(TXEN));
  UCSRC |= (_BV(UCSZ1) | _BV(UCSZ0));
  UBRRH = (129 >> 8);
  UBRRL = (129 & 0xff);

  /* LCD */
  lcd_init();

  /* |[g(gps̓vAbv) */
  PORTA |= (_BV(PA1));
  PORTB |= (_BV(PB3) | _BV(PB4) | _BV(PB6));
  PORTD |= (_BV(PD2) | _BV(PD3) | _BV(PD4) | _BV(PD5));

  for (; ; ) {
    if ((Modified & _BV(MOD_GPS))) {
      display();
      Modified &= ~_BV(MOD_GPS);
    }
    if ((Modified & _BV(MOD_CAP))) {
      adjust();
      monitor();
      Modified &= ~_BV(MOD_CAP);
    }

    cli();
    if (Modified) continue;
    /* 荞݂҂(sleepmɎs悤ɒOsei) */
    asm volatile(
      "sei\n\t"
      "sleep\n\t"
    ::);
  }

  return 0;
}


/*
  LCD֏\
*/
static void display() {
  memcpy(Lcd_buf, Gps_time, 6);
  Lcd_buf[6] = '0' + ((Gps_sat < 10) ? Gps_sat : 9);
  Lcd_buf[7] = '0' + ((Cap_err < 10) ? Cap_err : 9);
  if (Gps_stt == 0) {
    Lcd_buf[8] = 0x00;
  } else {
    Lcd_buf[8] = (0x40 >> Gate);
  }
  if (Gate == 2 && Cap_err <= 5) {
    Lcd_buf[8] |= 0x80;
  }
  lcd_refresh();

  return;
}


/*
  VCTCXO̒
*/
static void adjust() {
  static uint8_t lerr = 0, herr = 0;
  uint8_t mag;
  uint8_t data[3];
  uint8_t tmp;

  /* ʂ̌ */
  if (Gate == 0) {
    mag = 100;
  } else if (Gate == 1) {
    mag = 10;
  } else {
    mag = 1;
  }
  Adj_tune -= ADJ_K0 * Cap_dif * mag / ADJ_K1;

  /* VCTCXO̎gƉx̎擾 */
  data[0] = 0x00;
  tmp = ((uint16_t)Adj_tune >> 8);
  data[1] = (((tmp & 0x80) >> 1) | (tmp & 0x3f));	/* l͈̔͂𒴂ȂƉ */
  data[2] = ((uint16_t)Adj_tune & 0xff);
  i2c_sendrecv(((0x41 << 1) | 0), 3, data, 0);
  data[0] = 0x02;
  i2c_sendrecv(((0x41 << 1) | 0), 1, data, 0);
  i2c_sendrecv(((0x41 << 1) | 1), 2, data, 1);
  Adj_temp = (((uint16_t)data[0] << 8) | data[1]);;

  /* Q[gԂ̐؂ւ */
  if (Cap_err <= ADJ_LTHR) lerr++; else lerr = 0;
  if (Cap_err >= ADJ_HTHR) herr++; else herr = 0;
  if (lerr >= ADJ_LCNT && Gate != 2) {
    Gate++;
    lerr = 0;
  }
  if (herr >= ADJ_HCNT && Gate != 0) {
    Gate--;
    herr = 0;
  }

  return;
}


/*
  PC֌݂̏Ԃ𑗐M
*/
static void monitor() {
  uart_putnum(Gate);
  uart_puts(Gps_time, 6);
  uart_putnum(Gps_stt);
  uart_putnum(Gps_sat);
  uart_putnum(Cap_out);
  uart_putnum(Cap_dif);
  uart_putnum(Adj_tune);
  uart_putnum(Adj_temp);
  uart_putc('\r');
  uart_putc('\n');

  return;
}


/*
  UART֕o
*/
static void uart_putc(uint8_t chr) {
  while (!(UCSRA & _BV(UDRE))) ;
  UDR = chr;

  return;
}


/*
  UART֕{󔒂o
*/
static void uart_puts(uint8_t *str, uint8_t len) {
  do {
    uart_putc(*str++);
  } while (--len);
  uart_putc(' ');

  return;
}


/*
  UART֐l{󔒂o
*/
static void uart_putnum(int16_t num) {
  int8_t i;
  uint16_t n;
  uint8_t buf[6];
  uint8_t minus;

  if (num < 0) {
    minus = 1;
    n = -num;
  } else {
    minus = 0;
    n = num;
  }

  for (i = 5; ; i--) {
    buf[i] = n % 10 + '0';
    n /= 10;
    if (n == 0) {
      if (minus) buf[--i] = '-';
      break;
    }
  }
  uart_puts(buf + i, 6 - i);

  return;
}


/*
  UARTM荞(GGAZeX̓ǂݍ)
*/
ISR (USART_RX_vect) {
  static uint8_t buf[BUF_SIZE];
  static uint8_t len = 0;

  buf[len++] = UDR;
  if (len != BUF_SIZE) {
    if (buf[len - 1] == '\n') len = 0;	/* GGA̐擪ȊOȂ̂ŃZbg */
    return;
  }
  len = 0;

  /* GGAǂ`FbN */
  if (strncmp((char *)buf, "$GPGGA", 6) != 0) return;

  /* f[^̎擾(LR9543̏o͂) */
  memcpy(Gps_time, buf + 7, 6);	/* (UTC->JST) */
  Gps_time[1] += 9;	/* UTC->JST */
  if (Gps_time[1] > '9') {
    Gps_time[1] -= 10;
    Gps_time[0]++;
    if (Gps_time[0] == '2' && Gps_time[1] >= '4') {
      Gps_time[0] -= 2;
      Gps_time[1] -= 4;
    }
  }
  Gps_stt = buf[43] - '0';
  Gps_sat = 10 * (buf[45] - '0') + buf[46] - '0';
  Modified |= _BV(MOD_GPS);

  return;
}


/*
  1ppsLv`荞
*/
ISR (TIMER1_CAPT_vect) {
  uint16_t cap;
  int16_t err;

  cap = ICR1;

  /* TvO̊Jn */
  if (Cap_new) {
    Cap_num = 0;
    Cap_dif = 0;
    Cap_out = 0;
    Cap_new = 0;
    Cap_cnt = -cap;

    return;
  }

  /* Lv`ꂽlWv */
  Cap_cnt += cap;
  err = (int16_t)((int32_t)Cap_cnt - OSC_FRQ);

  /* Ol͖ */
  if (Gps_stt == 0 || err < -OSC_TOL || OSC_TOL < err) {
    if (Cap_out < 255) Cap_out++;
    Cap_cnt = -cap;
    return;
  }

  /* Q[gԂ̓vXV */
  Cap_dif += err;
  Cap_num++;
  if ((Gate == 1 && Cap_num != 10) || (Gate == 2 && Cap_num != 100)) {
    Cap_cnt = -cap;
    return;
  }
  Cap_err = (Cap_dif < 0) ? -Cap_dif : Cap_dif;

  /* VTvOJn */
  Cap_new = 1;
  Modified |= _BV(MOD_CAP);

  return;
}


/*
  JE^I[o[t[荞
*/
ISR (TIMER1_OVF_vect) {
  Cap_cnt += 65536;

  return;
}
